/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package sf.recallEstimation;

import java.util.concurrent.Semaphore;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.apache.commons.math4.analysis.UnivariateFunction;
import org.apache.commons.math4.analysis.integration.BaseAbstractUnivariateIntegrator;
import sf.recallEstimation.UnivariateFunctionImpl.MuB;
import sf.recallEstimation.UnivariateFunctionImpl.PrRankXLessM;
import sf.recallEstimation.UnivariateFunctionImpl.Sigma2BThread;
import sf.recallEstimation.integrating.OnePeakPositiveFunctionIntegrator;

/**
 *
 * @author xmic
 */
public class RecallEstimator {

    public static final int PARALELIZATION = 8;
    private final int nDatasetSize;
    private final int k;
    private final double maxDistance;
    private final Double sketchesIDim;
    private final UnivariateFunction fxDistDistr;
    private final UnivariateFunction pxProbSkBitsDif;
    private final int bSkLength;
    private final double x0;
    private final double[] mub;
    private final double[] sigmab;

    public RecallEstimator(int nDatasetSize, int k, double maxDistance, Double sketchesIDim, int bSkLength, UnivariateFunction fxDistDistr, UnivariateFunction pxProbSkBitsDif) {
        this.nDatasetSize = nDatasetSize;
        this.k = k;
        this.maxDistance = maxDistance;
        this.fxDistDistr = fxDistDistr;
        this.bSkLength = bSkLength;
        this.pxProbSkBitsDif = pxProbSkBitsDif;
        this.sketchesIDim = sketchesIDim;
        x0 = estimateX0();
        mub = evaluateMuBArray();
//        sigmab = evaluateSigmaBArray(PARALELIZATION);
        sigmab = evaluateSigmaBArrayAccordingToMubArray();
    }

    public RecallEstimator(int nDatasetSize, int k, double maxDistance, Double sketchesIDim, int bSkLength, UnivariateFunction fxDistDistr, UnivariateFunction pxProbSkBitsDif, double[] mub, double[] sigmab, boolean scaleMubSigmab) {
        this.nDatasetSize = nDatasetSize;
        this.k = k;
        this.maxDistance = maxDistance;
        this.fxDistDistr = fxDistDistr;
        this.bSkLength = bSkLength;
        this.pxProbSkBitsDif = pxProbSkBitsDif;
        this.sketchesIDim = sketchesIDim;
        x0 = estimateX0();
        if(scaleMubSigmab){
            double coef = nDatasetSize / mub[mub.length - 1];
            for (int i = 0; i < mub.length; i++) {
                mub[i] = mub[i] * coef;
                sigmab[i] = sigmab[i] * coef;                
            }
        }
        this.mub = mub;
        this.sigmab = sigmab;
    }

    public double estimateRecall(int mCandidateSetSize) {
        RecallIntegralBody recallIntegralBody = new RecallIntegralBody(mub, sigmab, mCandidateSetSize);
        BaseAbstractUnivariateIntegrator integrator = new OnePeakPositiveFunctionIntegrator(x0);
        double integral = integrator.integrate(Integer.MAX_VALUE, recallIntegralBody, 0, x0);
        return nDatasetSize / k * integral;
    }

    private double estimateX0() {
        double curX0 = maxDistance / 2;
        double curStep = maxDistance / 4;
        double maxDev = 1E-8;
        double curDev = Double.MAX_VALUE;
        BaseAbstractUnivariateIntegrator integrator = new OnePeakPositiveFunctionIntegrator(0, maxDistance);
        while (Math.abs(curDev) > maxDev) {
            double curK = nDatasetSize * integrator.integrate(Integer.MAX_VALUE, fxDistDistr, 0, curX0);
//            Logger.getLogger(RecallEstimator.class.getName()).log(Level.INFO, "Estimating x0. Current k: {0}, goal: {1}, limit of deviance: {2}, deviance: {3}. X_0: {4}", new Object[]{curK, k, Double.toString(maxDev), Double.toString(curDev), curX0});
            curDev = curK - k;
            if (curDev < 0) {
                curX0 += curStep;
            } else {
                curX0 -= curStep;
            }
            curStep = curStep / 2;
        }
        Logger.getLogger(RecallEstimator.class.getName()).log(Level.INFO, "Finished evaluating x0: {0}", Double.toString(curX0));
        return curX0;
    }

    private double[] evaluateMuBArray() {
        MuB mubEvaluator = new MuB(nDatasetSize, bSkLength, maxDistance, sketchesIDim, fxDistDistr, pxProbSkBitsDif);
        double[] ret = new double[bSkLength + 1];
        for (int i = 0; i <= bSkLength; i++) {
            ret[i] = mubEvaluator.evaluateMuB(i);
        }
        Logger.getLogger(RecallEstimator.class.getName()).log(Level.INFO, "Finished evaluating mean values.");
        return ret;
    }

    private double[] evaluateSigma2BArray(int paralelization) {
        double[] ret = new double[bSkLength + 1];
        Semaphore sem = new Semaphore(paralelization);
        Sigma2BThread[] thds = new Sigma2BThread[bSkLength + 1];
        AtomicInteger counter = new AtomicInteger(0);
        for (int i = bSkLength; i >= 0; i--) {
            thds[i] = new Sigma2BThread(counter, ret.length, nDatasetSize, i, bSkLength, maxDistance, sketchesIDim, fxDistDistr, pxProbSkBitsDif, sem);
            try {
                sem.acquire();
                new Thread(thds[i]).start();
            } catch (InterruptedException ex) {
                Logger.getLogger(RecallEstimator.class.getName()).log(Level.SEVERE, null, ex);
            }
        }
        for (int i = 0; i < paralelization; i++) {
            try {
                sem.acquire();
            } catch (InterruptedException ex) {
                Logger.getLogger(RecallEstimator.class.getName()).log(Level.SEVERE, null, ex);
            }
        }
        for (int i = 0; i <= bSkLength; i++) {
            ret[i] = thds[i].getResult();
        }
        Logger.getLogger(RecallEstimator.class.getName()).log(Level.INFO, "Finished evaluating variances.");
        return ret;
    }

    private double[] evaluateSigmaBArray(int paralelization) {
        double[] sigma2BArray = evaluateSigma2BArray(paralelization);
        double[] ret = new double[sigma2BArray.length];
        for (int i = 0; i < ret.length; i++) {
            ret[i] = Math.sqrt(sigma2BArray[i]);
        }
        return ret;
    }

    private double[] evaluateSigmaBArrayAccordingToMubArray() {
        double[] ret = new double[mub.length];
        ret[0] = mub[0] / 2;
        for (int i = 1; i < mub.length; i++) {
            ret[i] = (mub[i] - mub[i - 1]) / 2 - ret[i - 1];
            Logger.getLogger(RecallEstimator.class.getName()).log(Level.INFO, "Variance for {0} bits out of {1} evaluated. Result: {2}", new Object[]{i + 1, mub.length, Double.toString(ret[i])});
            if (ret[i] < 0) {
                Logger.getLogger(RecallEstimator.class.getName()).log(Level.WARNING, "Variance for {0} bit is negative: {1}. Was replaced with absolute value.", new Object[]{i + 1, Double.toString(ret[i])});
                ret[i] = Math.abs(ret[i]);
            }
            if (ret[i] == 0) {
                Logger.getLogger(RecallEstimator.class.getName()).log(Level.WARNING, "Variance for {0} bit is zero. Was replaced with min double value.", new Object[]{i + 1});
                ret[i] = Double.MIN_VALUE;
            }
        }
        return ret;
    }

    private class RecallIntegralBody implements UnivariateFunction {

        private final PrRankXLessM prRankXLessM;

        public RecallIntegralBody(double[] mub, double[] sigmab, int mCandidateSetSize) {
            this.prRankXLessM = new PrRankXLessM(sketchesIDim, pxProbSkBitsDif, bSkLength, mCandidateSetSize, mub, sigmab);
        }

        @Override
        public double value(double x) {
            double prRank = prRankXLessM.value(x);
            double fx = fxDistDistr.value(x);
            return prRank * fx;
        }

    }

}
